//
// Syd: rock-solid unikernel
// src/proc.rs: DNS utilities
//
// Copyright (c) 2024 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0

use std::{
    ffi::CString,
    net::{IpAddr, Ipv4Addr, Ipv6Addr},
    ptr,
};

use indexmap::IndexSet;
use nix::{
    errno::Errno,
    libc::{
        addrinfo, freeaddrinfo, getaddrinfo, sockaddr_in, sockaddr_in6, AF_INET, AF_INET6,
        AF_UNSPEC, EAI_AGAIN, EAI_BADFLAGS, EAI_FAIL, EAI_FAMILY, EAI_MEMORY, EAI_NONAME,
        EAI_SERVICE, EAI_SOCKTYPE, EAI_SYSTEM,
    },
};

use crate::fs::fillrandom;

/// Resolves a hostname to a single IP address.
/// In case of multiple responses, an IP is selected randomly.
/// Randomness is provided by getrandom(2).
pub fn resolve_rand(name: &str, family: Option<i32>) -> Result<IpAddr, Errno> {
    // Read random bytes with getrandom(2) and convert to usize.
    // Note, getrandom(2) is soon to be in the VDSO!
    let mut buf = [0u8; 4];
    if fillrandom(&mut buf).is_err() {
        return Err(Errno::EIO); // Input/output error.
    }
    // Convert bytes to a usize.
    let cookie = usize::try_from(u32::from_ne_bytes(buf)).unwrap_or(usize::MAX);

    // Resolve hostname.
    let addrs = resolve_host(name, family)?;

    // Select a random IP address from the list.
    // Calculate random index within the bounds of the addresses vector.
    #[allow(clippy::arithmetic_side_effects)]
    Ok(addrs[cookie.wrapping_rem(addrs.len())])
}

/// Resolves a hostname using the system DNS resolver.
pub fn resolve_host(name: &str, family: Option<i32>) -> Result<Vec<IpAddr>, Errno> {
    // Convert the hostname to a CString, ensuring it doesn't contain null bytes.
    let c_name = match CString::new(name) {
        Ok(cstring) => cstring,
        Err(_) => return Err(Errno::EINVAL), // Invalid argument if string contains null byte.
    };

    let ai_family = match family {
        Some(AF_INET) => AF_INET,
        Some(AF_INET6) => AF_INET6,
        Some(_) => return Err(Errno::EINVAL),
        _ => AF_UNSPEC, // Allow IPv4 or IPv6.
    };

    // Prepare hints for getaddrinfo.
    let hints = addrinfo {
        ai_family,
        ai_flags: 0,    // No special flags.
        ai_socktype: 0, // Any type.
        ai_protocol: 0, // Any protocol.
        ai_addrlen: 0,
        ai_canonname: ptr::null_mut(),
        ai_addr: ptr::null_mut(),
        ai_next: ptr::null_mut(),
    };

    // Result pointer for getaddrinfo.
    let mut res: *mut addrinfo = ptr::null_mut();

    // SAFETY: Call getaddrinfo to resolve the hostname.
    let ret = unsafe { getaddrinfo(c_name.as_ptr(), ptr::null(), &hints, &mut res) };

    if ret != 0 {
        // getaddrinfo failed.
        if ret == EAI_SYSTEM {
            // System error occurred; errno is set.
            return Err(Errno::last());
        } else {
            // Map getaddrinfo error codes to Errno.
            let err = match ret {
                EAI_AGAIN => Errno::EAGAIN,             // Temporary failure.
                EAI_BADFLAGS => Errno::EINVAL,          // Invalid value.
                EAI_FAIL => Errno::EIO,                 // Non-recoverable error.
                EAI_FAMILY => Errno::EAFNOSUPPORT,      // Address family not supported.
                EAI_MEMORY => Errno::ENOMEM,            // Memory allocation failure.
                EAI_NONAME => Errno::ENOENT,            // Name or service not known.
                EAI_SERVICE => Errno::EPROTONOSUPPORT,  // Service not supported.
                EAI_SOCKTYPE => Errno::ESOCKTNOSUPPORT, // Socket type not supported.
                _ => Errno::EIO,                        // Generic I/O error.
            };
            return Err(err);
        }
    }

    // Create an IndexSet to store unique IPs while preserving insertion order.
    let mut addrs = IndexSet::new();

    // Traverse the linked list of addrinfo structures.
    let mut ptr = res;
    while !ptr.is_null() {
        // SAFETY: getaddrinfo returned success.
        // In libc, we trust.
        unsafe {
            let ai = &*ptr;
            if !ai.ai_addr.is_null() {
                let sockaddr = &*(ai.ai_addr as *const libc::sockaddr);
                // Extract the IP address based on the family.
                #[allow(clippy::cast_ptr_alignment)]
                match i32::from(sockaddr.sa_family) {
                    AF_INET if matches!(ai_family, AF_INET | AF_UNSPEC) => {
                        // IPv4 address.
                        let addr_in = &*(ai.ai_addr as *const sockaddr_in);
                        let ip = IpAddr::V4(Ipv4Addr::from(u32::from_be(addr_in.sin_addr.s_addr)));
                        addrs.insert(ip);
                    }
                    AF_INET6 if matches!(ai_family, AF_INET6 | AF_UNSPEC) => {
                        // IPv6 address.
                        let addr_in6 = &*(ai.ai_addr as *const sockaddr_in6);
                        let ip = IpAddr::V6(Ipv6Addr::from(addr_in6.sin6_addr.s6_addr));
                        addrs.insert(ip);
                    }
                    _ => {
                        // Unknown address family; ignore.
                    }
                }
            }
            ptr = ai.ai_next;
        }
    }

    // SAFETY: Free the addrinfo linked list allocated by getaddrinfo.
    unsafe {
        freeaddrinfo(res);
    }

    if addrs.is_empty() {
        // No addresses were found.
        Err(Errno::ENOENT) // No such file or directory.
    } else {
        Ok(addrs.iter().copied().collect())
    }
}
