//
// Syd: rock-solid unikernel
// src/log.rs: Simple logging on standard error using JSON lines
//
// Copyright (c) 2023, 2024 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0

/* Simple logging with JSON lines */
use std::{
    env,
    ffi::CString,
    fs::{File, OpenOptions},
    io::{self, Write},
    os::{
        fd::{FromRawFd, RawFd},
        unix::{ffi::OsStrExt, fs::OpenOptionsExt},
    },
    path::Path,
    sync::atomic::{AtomicBool, Ordering},
    time::{SystemTime, UNIX_EPOCH},
};

use btoi::btoi;
use hex::DisplayHex;
use log::{Level, LevelFilter, Log};
use nix::{
    libc::{STDERR_FILENO, STDIN_FILENO},
    unistd::{isatty, Pid, Uid},
};
use once_cell::sync::Lazy;
use parking_lot::Mutex;
use serde_json::{Map, Value};

use crate::{config::*, err::SydResult, proc::*};

pub(crate) static NO_SYSLOG: Lazy<bool> = Lazy::new(|| std::env::var_os(ENV_NO_SYSLOG).is_some());

/// info! logging macro
#[macro_export]
macro_rules! info {
    ($($key:literal : $value:expr),+) => {
        if log::log_enabled!(log::Level::Info) {
            let timestamp = $crate::log::now();
            let mut map = serde_json::Map::new();
            let syd = {
                use lexis::ToName;
                let pid = nix::unistd::Pid::this().as_raw();
                let tid = nix::unistd::gettid().as_raw();
                #[allow(clippy::cast_sign_loss)]
                let lex = (tid as u64).to_name();
                if pid == tid {
                    format!("{lex}@{pid}")
                } else {
                    format!("{lex}@{tid}/{pid}")
                }
            };
            map.insert("syd".to_string(), serde_json::Value::String(syd));
            $(
                if let Ok(value) = serde_json::to_value($value) {
                    map.insert($key.to_string(), value);
                } else {
                    map.insert($key.to_string(), serde_json::Value::Null);
                }
            )+
            if !map.is_empty() {
                $crate::log::log(log::Level::Info, timestamp, map);
            }
        }
    }
}

/// error! logging macro
#[macro_export]
macro_rules! error {
    ($($key:literal : $value:expr),+) => {
        if log::log_enabled!(log::Level::Error) {
            let timestamp = $crate::log::now();
            let mut map = serde_json::Map::new();
            let syd = {
                use lexis::ToName;
                let pid = nix::unistd::Pid::this().as_raw();
                let tid = nix::unistd::gettid().as_raw();
                #[allow(clippy::cast_sign_loss)]
                let lex = (tid as u64).to_name();
                if pid == tid {
                    format!("{lex}@{pid}")
                } else {
                    format!("{lex}@{tid}/{pid}")
                }
            };
            map.insert("syd".to_string(), serde_json::Value::String(syd));
            $(
                if let Ok(value) = serde_json::to_value($value) {
                    map.insert($key.to_string(), value);
                } else {
                    map.insert($key.to_string(), serde_json::Value::Null);
                }
            )+
            if !map.is_empty() {
                $crate::log::log(log::Level::Error, timestamp, map);
            }
        }
    }
}

/// warn! logging macro
#[macro_export]
macro_rules! warn {
    ($($key:literal : $value:expr),+) => {
        if log::log_enabled!(log::Level::Warn) {
            let timestamp = $crate::log::now();
            let mut map = serde_json::Map::new();
            let syd = {
                use lexis::ToName;
                let pid = nix::unistd::Pid::this().as_raw();
                let tid = nix::unistd::gettid().as_raw();
                #[allow(clippy::cast_sign_loss)]
                let lex = (tid as u64).to_name();
                if pid == tid {
                    format!("{lex}@{pid}")
                } else {
                    format!("{lex}@{tid}/{pid}")
                }
            };
            map.insert("syd".to_string(), serde_json::Value::String(syd));
            $(
                if let Ok(value) = serde_json::to_value($value) {
                    map.insert($key.to_string(), value);
                } else {
                    map.insert($key.to_string(), serde_json::Value::Null);
                }
            )+
            if !map.is_empty() {
                $crate::log::log(log::Level::Warn, timestamp, map);
            }
        }
    }
}

/// debug! logging macro
#[macro_export]
macro_rules! debug {
    ($($key:literal : $value:expr),+) => {
        if log::log_enabled!(log::Level::Debug) {
            let timestamp = $crate::log::now();
            let mut map = serde_json::Map::new();
            $(
                if let Ok(value) = serde_json::to_value($value) {
                    map.insert($key.to_string(), value);
                } else {
                    map.insert($key.to_string(), serde_json::Value::Null);
                }
            )+
            if !map.is_empty() {
                $crate::log::log(log::Level::Debug, timestamp, map);
            }
        }
    }
}

/// trace! logging macro
#[macro_export]
macro_rules! trace {
    ($($key:literal : $value:expr),+) => {
        if log::log_enabled!(log::Level::Trace) {
            let timestamp = $crate::log::now();
            let mut map = serde_json::Map::new();
            $(
                if let Ok(value) = serde_json::to_value($value) {
                    map.insert($key.to_string(), value);
                } else {
                    map.insert($key.to_string(), serde_json::Value::Null);
                }
            )+
            if !map.is_empty() {
                $crate::log::log(log::Level::Trace, timestamp, map);
            }
        }
    }
}

type Writer = Box<dyn Write + Send>;

// Is the log file descriptor a TTY?
static LOG_TTY: AtomicBool = AtomicBool::new(false);

/// Simple logging on standard error using JSON lines
pub struct JsonLinesLogger {
    filter: LevelFilter,
    writer: Mutex<Writer>,
}

impl JsonLinesLogger {
    fn new(log: Option<&Path>, default_level: &str) -> SydResult<Self> {
        let level = env::var_os(ENV_LOG)
            .map(|s| s.to_string_lossy().to_string())
            .unwrap_or_else(|| default_level.to_string());
        let filter = level.parse::<LevelFilter>().unwrap_or(LevelFilter::Warn);

        let mut tty = false;
        let writer: Box<dyn io::Write + Send> = if let Some(log) = log {
            Box::new(
                OpenOptions::new()
                    .append(true)
                    .create(true)
                    .mode(0o600)
                    .custom_flags(nix::libc::O_CLOEXEC)
                    .open(log)?,
            )
        } else {
            match env::var_os(ENV_LOG_FD) {
                None => {
                    // Check for TTY if logging to standard error.
                    // This allows for pretty printing logs.
                    tty = env::var_os(ENV_LOG).is_none()
                        && isatty(STDIN_FILENO).unwrap_or(false)
                        && isatty(STDERR_FILENO).unwrap_or(false);
                    Box::new(io::stderr())
                }
                Some(fd) => {
                    // SAFETY: We panic on parse errors.
                    #[allow(clippy::disallowed_methods)]
                    let fd = btoi::<RawFd>(fd.as_os_str().as_bytes())
                        .expect("Invalid log file descriptor");
                    // SAFETY: We trust the user to pass a valid FD.
                    Box::new(unsafe { File::from_raw_fd(fd) })
                }
            }
        };

        // Record LOG_TTY information.
        LOG_TTY.store(tty, Ordering::Relaxed);

        Ok(Self {
            filter,
            writer: Mutex::new(writer),
        })
    }

    /// Initialize logging
    pub fn init(log: Option<&Path>, default_level: &str) -> SydResult<()> {
        let logger = Self::new(log, default_level)?;

        log::set_max_level(logger.filter);
        Ok(log::set_boxed_logger(Box::new(logger))?)
    }
}

impl Log for JsonLinesLogger {
    fn enabled(&self, metadata: &log::Metadata) -> bool {
        metadata.level() <= self.filter
    }

    fn log(&self, record: &log::Record) {
        if self.enabled(record.metadata()) {
            let mut writer = self.writer.lock();
            let _ = writeln!(writer, "{}", record.args());
        }
    }

    fn flush(&self) {}
}

/// Helper for logging using JSON lines.
#[allow(clippy::cognitive_complexity)]
pub fn log(level: Level, timestamp: u64, mut msg: Map<String, Value>) {
    // Check if the log output is a TTY for info, warn and error.
    let add = level <= Level::Warn;
    let tty = add && LOG_TTY.load(Ordering::Relaxed);

    // Check if the log message has process ID info.
    if let Some(pid) = msg.get("pid").and_then(|v| v.as_i64()) {
        // Process ID is given.
        // If TTY or log feature is enabled enrichen the message with cmdline and cwd.
        // If !TTY or log feature is not enabled enrichen the message with comm only.
        // SAFETY: proc_{comm,cmdline,cwd} returns an XPath whose Display implementation
        // masks control characters.
        #[allow(clippy::cast_possible_truncation)]
        let pid = Pid::from_raw(pid as nix::libc::pid_t);
        if pid.as_raw() != 0 {
            if tty {
                if let Ok(cmd) = proc_cmdline(pid) {
                    msg.insert("cmd".to_string(), Value::String(cmd.to_string()));
                } else {
                    msg.insert("cmd".to_string(), Value::Null);
                }
            } else if let Ok(cmd) = proc_comm(pid) {
                msg.insert("cmd".to_string(), Value::String(cmd.to_string()));
            } else {
                msg.insert("cmd".to_string(), Value::Null);
            }
            if let Ok(dir) = proc_cwd(pid) {
                msg.insert("cwd".to_string(), Value::String(dir.to_string()));
            } else {
                msg.insert("cwd".to_string(), Value::Null);
            }
        }
    }

    // Reorder process id if TTY for readability.
    if tty {
        if let Some(pid) = msg.remove("pid") {
            msg.insert("pid".to_string(), pid);
        }
    }

    // Enrich the log message with the current user.
    if add {
        msg.insert(
            "uid".to_string(),
            Value::Number(Uid::current().as_raw().into()),
        );
    }

    // Enrich the log message with an ISO8601 format timestamp.
    msg.insert("time".to_string(), Value::String(format_iso8601(timestamp)));

    // Reorder message fields to the end if TTY for readability.
    if tty {
        if let Some(src) = msg.remove("src") {
            msg.insert("src".to_string(), src);
        }
        if let Some(err) = msg.remove("error") {
            msg.insert("error".to_string(), err);
        }
        if let Some(m) = msg.remove("msg") {
            msg.insert("msg".to_string(), m);
        }
        if let Some(tip) = msg.remove("tip") {
            msg.insert("tip".to_string(), tip);
        }
    }

    // Serialize the message into a String.
    let msg_data = serde_json::to_string(&msg).unwrap_or_else(|e| {
        let e = serde_json::to_string(&format!("{e:?}")).unwrap_or("?".to_string());
        format!("{{\"ctx\":\"log\",\"op\":\"serialize\",\"error\": \"{e}\"}}")
    });
    let msg_info = if level == Level::Warn && LOG_TTY.load(Ordering::Relaxed) {
        Some(serde_json::to_string_pretty(&msg).unwrap_or_else(|e| {
            let e = serde_json::to_string(&format!("{e:?}")).unwrap_or("?".to_string());
            format!("{{\"ctx\":\"log\",\"op\":\"serialize\",\"error\": \"{e}\"}}")
        }))
    } else {
        None
    };
    let msg_info = msg_info.as_ref().unwrap_or(&msg_data);

    // Send the log String to its destination.
    match level {
        Level::Error => {
            if !*NO_SYSLOG {
                syslog(&msg_data);
            }
            log::error!("{msg_data}");
        }
        Level::Warn => {
            if !*NO_SYSLOG {
                syslog(&msg_data);
            }
            log::warn!("{msg_info}");
        }
        Level::Info => log::info!("{msg_data}"),
        Level::Debug => {
            crate::t(&msg_data.to_string());
            log::debug!("{msg_data}");
        }
        Level::Trace => crate::t(&msg_data.to_string()),
    }
}

/// Logs a message to the system's syslog.
///
/// # Arguments
///
/// * `message` - The message string to be logged.
fn syslog(message: &str) {
    let c_msg = CString::new(message).unwrap_or_else(|_|
        // SAFETY: We ensure the string has no null bytes and the vector
        // we pass into the function is an actual byte vector.
        unsafe { CString::from_vec_unchecked(b"?".to_vec()) });

    // SAFETY: Use the syslog interface provided by libc.
    unsafe {
        nix::libc::syslog(
            nix::libc::LOG_WARNING,
            "%s\0".as_ptr().cast(),
            c_msg.as_ptr(),
        );
    }
}

/// Return the current time in seconds since the unix epoch.
pub fn now() -> u64 {
    SystemTime::now()
        .duration_since(UNIX_EPOCH)
        .unwrap_or_default()
        .as_secs()
}

/// Returns current time as a compact ISO8601-formatted string.
///
/// The format is currently "YYYYMMDDThhmmssZ". The format may change in
/// the future but it will always remain conformant to the ISO8601
/// standard.
#[allow(clippy::arithmetic_side_effects)]
#[allow(deprecated)]
fn format_iso8601(timestamp: u64) -> String {
    let timestamp = match nix::libc::time_t::try_from(timestamp) {
        Ok(t) => t,
        Err(_) => return "error on time()".to_string(),
    };
    let mut tm = std::mem::MaybeUninit::uninit();

    // SAFETY: gmtime_r returns NULL on error.
    if unsafe { nix::libc::gmtime_r(&timestamp, tm.as_mut_ptr()) }.is_null() {
        return "error on gmtime_r()".to_string();
    }

    // SAFETY: gmtime_r returned success.
    let tm = unsafe { tm.assume_init() };

    let year = tm.tm_year + 1900;
    // tm_mon ranges from 0 to 11.
    let month = tm.tm_mon + 1;
    let day = tm.tm_mday;
    let hour = tm.tm_hour;
    let minute = tm.tm_min;
    let second = tm.tm_sec;

    // Format: YYYYMMDDThhmmssZ
    format!("{year:04}{month:02}{day:02}T{hour:02}{minute:02}{second:02}Z")
}

/// Logs an untrusted buffer, escaping it as hex if it contains control characters.
/// Returns a boolean in addition to the String which is true if String is hex-encoded.
pub fn log_untrusted_buf(buf: &[u8]) -> (String, bool) {
    if contains_ascii_unprintable(buf) {
        (buf.to_upper_hex_string(), true)
    } else if let Ok(s) = std::str::from_utf8(buf) {
        (s.to_string(), false)
    } else {
        (buf.to_upper_hex_string(), true)
    }
}

/// Checks if the buffer contains ASCII unprintable characters.
pub fn contains_ascii_unprintable(buf: &[u8]) -> bool {
    buf.iter().any(|byte| !is_ascii_printable(*byte))
}

/// Checks if the given character is ASCII printable.
pub fn is_ascii_printable(byte: u8) -> bool {
    (0x20..=0x7e).contains(&byte)
}
