//
// Syd: rock-solid unikernel
// src/pool.rs: Self growing / shrinking `ThreadPool` implementation
//
// Copyright (c) 2024 Ali Polatel <alip@chesswob.org>
// Based in part upon rusty_pool which is:
//     Copyright (c) Robin Friedli <robinfriedli@icloud.com>
//     SPDX-License-Identifier: Apache-2.0
//
// SPDX-License-Identifier: GPL-3.0

// Last sync with rusty_pool:
// Version 0.7.0
// Commit:d56805869ba3cbe47021d5660bbaf19ac5ec4bfb

use std::{
    mem::MaybeUninit,
    option::Option,
    os::fd::{AsRawFd, BorrowedFd, FromRawFd, OwnedFd, RawFd},
    sync::{
        atomic::{AtomicBool, AtomicU32, Ordering},
        Arc,
    },
    thread,
    thread::JoinHandle,
};

use libseccomp::{scmp_cmp, ScmpAction, ScmpFilterContext, ScmpSyscall};
use libseccomp_sys::{
    seccomp_notif, seccomp_notif_resp, seccomp_notify_receive, seccomp_notify_respond,
};
use nix::{
    errno::Errno,
    fcntl::{splice, tee, OFlag, SpliceFFlags},
    poll::PollTimeout,
    sys::epoll::{Epoll, EpollCreateFlags, EpollEvent, EpollFlags},
    unistd::{close, lseek64, pipe2, write, Whence},
};
use parking_lot::RwLock;
use rustc_hash::FxHashMap;

use crate::{
    cache::{AddrCache, PathCache},
    compat::{epoll_ctl_safe, ftruncate64},
    config::*,
    err::SydResult,
    error,
    fs::{lock_fd, retry_on_eintr, seal_memfd},
    hash::{
        aes_ctr_enc, aes_ctr_init, hmac_sha256_feed, hmac_sha256_fini, hmac_sha256_init,
        BLOCK_SIZE, HMAC_TAG_SIZE, IV, IV_SIZE,
    },
    hook::{HandlerMap, UNotifyEventRequest},
    info,
    path::XPathBuf,
    proc::proc_mmap,
    sandbox::{Capability, Sandbox},
    syscall_priority, ScmpNotifReq, SydArch, Sydcall,
};

const EOWNERDEAD: i32 = -nix::libc::EOWNERDEAD;

/// The absolute maximum number of workers. This corresponds to the
/// maximum value that can be stored within half the bits of u32, as two
/// counters (total workers and busy workers) are stored in one
/// AtomicU32.
pub(crate) const MAX_SIZE: u16 = u16::MAX;

/// Self growing / shrinking `ThreadPool` implementation.
#[derive(Clone)]
pub(crate) struct ThreadPool {
    core_size: u16,
    keep_alive: u16,
    safe_setid: bool,
    fd: RawFd,
    epoll: Arc<Epoll>,
    cache: Arc<(PathCache, AddrCache)>,
    sandbox: Arc<RwLock<Sandbox>>,
    handlers: Arc<HandlerMap>,
    crypt_map: Option<AesMap>,
    should_exit: Arc<AtomicBool>,
    worker_data: Arc<WorkerData>,
}

impl ThreadPool {
    /// Construct a new `ThreadPool` with the specified core pool size,
    /// max pool size and keep_alive time for non-core threads. This
    /// function creates an epoll instance and adds the seccomp fd to it
    /// but it does not spawn any threads.
    ///
    /// `core_size` specifies the amount of threads to keep alive for as
    /// long as the `ThreadPool` exists and the seccomp fd remains open.
    ///
    /// `keep_alive` specifies the duration in milliseconds for which to
    /// keep non-core pool worker threads alive while they do not
    /// receive any work.
    pub(crate) fn new(
        safe_setid: bool,
        core_size: u16,
        keep_alive: u16,
        fd: RawFd,
        cache: Arc<(PathCache, AddrCache)>,
        sandbox: Arc<RwLock<Sandbox>>,
        handlers: Arc<HandlerMap>,
        crypt_map: Option<AesMap>,
    ) -> Result<Self, Errno> {
        // Create epoll instance.
        let epoll = Epoll::new(EpollCreateFlags::EPOLL_CLOEXEC)?;

        // SAFETY: Borrow FD to make I/O safe API hippie.
        let seccomp_fd = unsafe { BorrowedFd::borrow_raw(fd) };

        // Add seccomp fd to epoll (becomes readable when system call is interrupted).
        // We do not need the data field as we operate on a single fd.
        // Quoting: https://idea.popcount.org/2017-02-20-epoll-is-fundamentally-broken-12/
        // """
        // The best and the only scalable approach is to use recent
        // Kernel 4.5+ and use level-triggered events with
        // EPOLLEXCLUSIVE flag. This will ensure only one thread is
        // woken for an event, avoid "thundering herd" issue and scale
        // properly across multiple CPU's.
        // """
        let event = EpollEvent::new(EpollFlags::EPOLLIN | EpollFlags::EPOLLEXCLUSIVE, 0);
        epoll_ctl_safe(&epoll, &seccomp_fd, Some(event))?;

        Ok(Self {
            fd,
            cache,
            sandbox,
            crypt_map,
            handlers,
            core_size,
            keep_alive,
            safe_setid,
            epoll: Arc::new(epoll),
            should_exit: Arc::new(AtomicBool::new(false)),
            worker_data: Arc::new(WorkerData::default()),
        })
    }

    /// Boot the thread pool. This is the main entry point.
    pub(crate) fn boot(self) -> Result<Option<JoinHandle<()>>, std::io::Error> {
        // Export seccomp rules if requested.
        // We have to prepare the filter twice if exporting,
        // as we cannot move it safely between threads...
        if std::env::var("SYD_SECX").is_ok() {
            println!("# Syd monitor rules");
            if let Ok(ctx) = Worker::prepare_confine(self.safe_setid) {
                let _ = ctx.export_pfc(&mut std::io::stdout());
            }

            println!("# Syd encryptor rules");
            if let Ok(ctx) = AesWorker::prepare_confine(self.safe_setid) {
                let _ = ctx.export_pfc(&mut std::io::stdout());
            }
        }

        // Spawn the AES thread if encryption is on.
        let crypt = {
            let sandbox = self.sandbox.read();
            if sandbox.enabled(Capability::CAP_CRYPT) {
                let crypt_fds = sandbox.crypt_setup().unwrap();
                let crypt_tmp = sandbox.crypt_tmp.clone();
                Some((crypt_fds, crypt_tmp))
            } else {
                None
            }
        };
        let crypt_handle = if let Some((crypt_fds, crypt_tmp)) = crypt {
            let crypt_map = self.crypt_map.as_ref().map(Arc::clone).unwrap();
            Some(
                self.try_spawn_aes(crypt_fds, crypt_map, crypt_tmp)
                    .expect("spawn AES encryption thread"),
            )
        } else {
            None
        };

        // Spawn the monitor thread which may confine itself, and spawn
        // emulator threads. Note, this will panic if it cannot spawn
        // the initial emulator thread which is going to tear everything
        // down.
        self.monitor()?;

        // Return join handle of the encryption thread,
        // so we can wait for ongoing encryption processes
        // before exiting the sandbox.
        Ok(crypt_handle)
    }

    /// Spawn a monitor thread that watches the worker pool busy count,
    /// and spawns new helper threads as necessary. This is done to
    /// ensure a sandbox process cannot DOS Syd by merely exhausting
    /// workers by e.g. opening the read end of a FIFO over and over
    /// again.
    #[allow(clippy::cognitive_complexity)]
    pub(crate) fn monitor(self) -> Result<(), std::io::Error> {
        thread::Builder::new()
            .name("syd_mon".to_string())
            .stack_size(MON_STACK_SIZE)
            .spawn(move || {
                // SAFETY: If sandbox is locked, confine right away.
                // Pass confined parameter to try_spawn so subsequent
                // spawned threads don't need to reapply the same filter
                // as it is inherited.
                let mut confined = if Sandbox::locked_once() {
                    Worker::prepare_confine(self.safe_setid)
                        .expect("prepare monitor thread confinement")
                        .load()
                        .expect("confine monitor thread");
                    info!("ctx": "confine_monitor_thread",
                        "msg": format!("monitor thread confined with{} SROP mitigation",
                            if self.safe_setid { "out" } else { "" }));
                    true
                } else {
                    false
                };
                let mut threads = vec![];
                // SAFETY: Panic if we cannot spawn the initial core thread.
                // There's little sense in continuing in this case.
                threads.push(
                    self.try_spawn(confined)
                        .expect("spawn initial emulator thread")
                        .unwrap(),
                );
                std::thread::sleep(MON_GRACE_TIME);
                loop {
                    // Confine as necessary.
                    if !confined && Sandbox::locked_once() {
                        Worker::prepare_confine(self.safe_setid)
                            .expect("prepare monitor thread confinement")
                            .load()
                            .expect("confine monitor thread");
                        info!("ctx": "confine_monitor_thread",
                            "msg": format!("monitor thread confined with{} SROP mitigation",
                                if self.safe_setid { "out" } else { "" }));
                        confined = true;
                    }
                    if self.should_exit.load(Ordering::Relaxed) {
                        break; // Time to exit.
                    } else if let Ok(Some(handle)) = self.try_spawn(confined) {
                        // When we spawned a new thread or there's an error spawning
                        // a thread we wait for the grace period.
                        threads.push(handle);
                        std::thread::sleep(MON_GRACE_TIME);
                    } else {
                        // When no idle threads are present we wait for cycle period.
                        std::thread::sleep(MON_CYCLE_TIME);
                    }
                    // Wait for finished threads.
                    let finished;
                    (finished, threads) = threads.into_iter().partition(|h| h.is_finished());
                    for handler in finished {
                        let _ = handler.join();
                    }
                }
            })
            .map(drop)
    }

    /// Try to create a new encryption thread.
    pub(crate) fn try_spawn_aes(
        &self,
        fdalg: (RawFd, RawFd),
        files: AesMap,
        tmpdir: Option<XPathBuf>,
    ) -> Result<JoinHandle<()>, std::io::Error> {
        AesWorker::new(fdalg, files, tmpdir, self.safe_setid).try_spawn()
    }

    /// Try to create a new worker thread as needed.
    /// Returns Ok(Some(JoinHandle)) if spawn succeeded, Ok(None) if no spawn was needed.
    pub(crate) fn try_spawn(
        &self,
        confined: bool,
    ) -> Result<Option<JoinHandle<()>>, std::io::Error> {
        // Create a new worker if there are no idle threads and the
        // current worker count is lower than the max pool size.
        let worker_count_val = self.worker_data.0.load(Ordering::Relaxed);
        let (curr_worker_count, busy_worker_count) = WorkerData::split(worker_count_val);

        let keep_alive = if curr_worker_count < self.core_size {
            // Create a new core worker if current pool size is below
            // core size during the invocation of this function.
            None
        } else if busy_worker_count < curr_worker_count {
            // We have idle threads, no need to spawn a new worker.
            return Ok(None);
        } else if curr_worker_count < MAX_SIZE {
            // Create a new helper worker if the current worker count is
            // below the MAX_SIZE and the pool has been observed to be
            // busy (no idle workers) during the invocation of this
            // function.
            Some(self.keep_alive)
        } else {
            // We cannot spawn anymore workers!
            // Ideally, this should never happen.
            return Ok(None);
        };

        // Try to spawn a new worker.
        Ok(Some(
            Worker::new(
                self.fd,
                self.safe_setid,
                Arc::clone(&self.epoll),
                Arc::clone(&self.cache),
                Arc::clone(&self.sandbox),
                Arc::clone(&self.handlers),
                keep_alive,
                Arc::clone(&self.should_exit),
                Arc::clone(&self.worker_data),
                self.crypt_map.as_ref().map(Arc::clone),
            )
            .try_spawn(confined)?,
        ))
    }
}

#[derive(Debug, Copy, Clone, Eq, PartialEq, Ord, PartialOrd)]
pub(crate) enum AesMod {
    Read,
    Append,
    Write,
}

impl From<OFlag> for AesMod {
    fn from(flags: OFlag) -> Self {
        if flags.contains(OFlag::O_WRONLY) || flags.contains(OFlag::O_RDWR) {
            if flags.contains(OFlag::O_APPEND) {
                Self::Append
            } else {
                Self::Write
            }
        } else {
            Self::Read
        }
    }
}

pub(crate) type AesMap = Arc<
    RwLock<(
        FxHashMap<XPathBuf, (RawFd, OwnedFd, AesMod, IV, u64, bool)>,
        bool,
    )>,
>;

#[derive(Clone)]
struct AesWorker {
    fdalg: (RawFd, RawFd),
    files: AesMap,
    tmpdir: Option<XPathBuf>,
    safe_setid: bool,
}

impl AesWorker {
    fn new(
        fdalg: (RawFd, RawFd),
        files: AesMap,
        tmpdir: Option<XPathBuf>,
        safe_setid: bool,
    ) -> Self {
        Self {
            fdalg,
            files,
            tmpdir,
            safe_setid,
        }
    }

    fn try_spawn(self) -> Result<JoinHandle<()>, std::io::Error> {
        thread::Builder::new()
            .name("syd_aes".to_string())
            .stack_size(AES_STACK_SIZE)
            .spawn(move || {
                // SAFETY: Logging will kill us after seccomp.
                info!("ctx": "confine_aes_thread",
                    "msg": format!("AES thread confined with{} SROP mitigation",
                        if self.safe_setid { "out" } else { "" }));
                // SAFETY: Default panic hook wont play well with seccomp
                std::panic::set_hook(Box::new(|_| {}));
                // SAFETY: Panic if we cannot confine the thread.
                Self::prepare_confine(self.safe_setid)
                    .unwrap()
                    .load()
                    .unwrap();
                // Enter main loop.
                Self::main(self.fdalg, self.files, self.tmpdir)
            })
    }

    fn main(setup_fds: (RawFd, RawFd), files: AesMap, tmpdir: Option<XPathBuf>) {
        let mut paths = vec![];
        let mut threads = vec![];
        loop {
            // Check if there're any pending encryption requests
            // for which we have not spawned a thread yet.
            // The thread is responsible for removing
            // the entry from the files map. This way
            // we ensure fstat() requests keep working
            // until we're done writing.
            let my_files = files.read();
            let finished = my_files.1;
            for (path, (crypt_fd, _, _, _, _, spawned)) in &my_files.0 {
                if !spawned {
                    // SAFETY: crypt_map keys are valid FDs.
                    let crypt_fd = unsafe { OwnedFd::from_raw_fd(*crypt_fd) };
                    paths.push((crypt_fd, path.clone()));
                }
            }
            drop(my_files);

            if paths.is_empty() {
                if finished {
                    // main-thread signaled exit.
                    break;
                } else {
                    // wait a bit and retry.
                    std::thread::sleep(AES_CYCLE_TIME);
                    continue;
                }
            }

            // Mark entries for which we're spawning a thread.
            let mut my_files = files.write();
            for (_, path) in &paths {
                if let Some(info) = my_files.0.get_mut(path) {
                    info.5 = true; // spawned=true.
                }
            }
            drop(my_files);

            for (crypt_fd, crypt_path) in paths.drain(..) {
                // Spawn a thread to handle the write.
                let files = Arc::clone(&files);
                threads.push(Self::spawn(
                    setup_fds,
                    crypt_fd,
                    crypt_path,
                    files,
                    tmpdir.is_none(),
                ));
            }

            // wait for a cycle.
            std::thread::sleep(AES_CYCLE_TIME);
        }

        // wait for the ongoing encryption operations before exiting.
        for thread in threads {
            thread.join().unwrap().unwrap();
        }
    }

    fn spawn(
        setup_fds: (RawFd, RawFd),
        crypt_fd: OwnedFd,
        crypt_path: XPathBuf,
        files: AesMap,
        memfd: bool,
    ) -> JoinHandle<Result<(), Errno>> {
        thread::Builder::new()
            .name("syd_aes".into())
            .stack_size(AES_STACK_SIZE)
            .spawn(move || {
                // Wait until we take a write lock on the encrypted fd.
                // This will succeed once all fds owned by the sandbox
                // process are closed.
                lock_fd(&crypt_fd, true, true)?;
                // All good, sync contents to disk.
                Self::sync(setup_fds, crypt_fd, crypt_path, files, memfd)
            })
            .unwrap()
    }

    #[allow(clippy::arithmetic_side_effects)]
    #[allow(clippy::cognitive_complexity)]
    fn sync(
        setup_fds: (RawFd, RawFd),
        crypt_fd: OwnedFd,
        crypt_path: XPathBuf,
        files: AesMap,
        memfd: bool,
    ) -> Result<(), Errno> {
        let (aes_fd, mac_fd) = setup_fds;

        let (_, enc_fd, file_mode, mut iv, _, _) = {
            let mut files = files.write();
            files.0.remove(&crypt_path).unwrap()
        };

        // Seal memfd to ensure no further writes happen.
        if memfd {
            seal_memfd(&crypt_fd).unwrap();
        }

        // Nothing to do if file was readonly.
        let mut is_append = match file_mode {
            AesMod::Read => return Ok(()),
            AesMod::Append => true,
            _ => false,
        };

        // Handle truncation quickly.
        #[allow(clippy::cast_sign_loss)]
        let data_size = lseek64(crypt_fd.as_raw_fd(), 0, Whence::SeekEnd).unwrap() as u64;
        if data_size == 0 {
            ftruncate64(enc_fd.as_raw_fd(), 0).unwrap();
            return Ok(());
        }

        // Handle opened for append but encrypted file is new.
        #[allow(clippy::cast_sign_loss)]
        let mut file_size = lseek64(enc_fd.as_raw_fd(), 0, Whence::SeekEnd)? as u64;
        if is_append && file_size == 0 {
            is_append = false;
        }

        // Handle opened for append but appended nothing quickly.
        if is_append
            && data_size
                <= file_size.saturating_sub((CRYPT_MAGIC.len() + HMAC_TAG_SIZE + IV_SIZE) as u64)
        {
            return Ok(());
        }

        // We handled quick cases, before possibly
        // truncating the encrypted file, let's
        // ensure we open the connections as expected.

        // Initialize HMAC socket and feed magic header and IV.
        let sock_mac = hmac_sha256_init(&mac_fd, false)?;
        hmac_sha256_feed(&sock_mac, CRYPT_MAGIC, true)?;
        hmac_sha256_feed(&sock_mac, iv.as_ref(), true)?;
        let (pipe_rd_mac, pipe_wr_mac) = pipe2(OFlag::O_CLOEXEC)?;

        // Handle last block re-encryption for append.
        if is_append {
            // Adjust file_size to exclude the header.
            let header_size = (CRYPT_MAGIC.len() + HMAC_TAG_SIZE + IV_SIZE) as u64;
            file_size -= header_size;

            // Calculate the offset of the last full block.
            let last_block_offset = if file_size % BLOCK_SIZE as u64 == 0 {
                file_size
            } else {
                file_size - (file_size % BLOCK_SIZE as u64)
            };

            // Adjust the IV counter based on the last full block offset.
            iv.add_counter(last_block_offset);

            // If there is a partial block at the end, we need to re-encrypt it.
            if last_block_offset < file_size {
                // Truncate the encrypted file to remove the partial block.
                let truncate_offset = header_size + last_block_offset;
                retry_on_eintr(|| {
                    ftruncate64(enc_fd.as_raw_fd(), truncate_offset.try_into().unwrap())
                })?;

                // Adjust crypt_fd to read from the last full block offset.
                #[allow(clippy::cast_possible_wrap)]
                lseek64(
                    crypt_fd.as_raw_fd(),
                    last_block_offset as i64,
                    Whence::SeekSet,
                )?;
            } else {
                // No partial block, start reading from the current file size.
                #[allow(clippy::cast_possible_wrap)]
                lseek64(crypt_fd.as_raw_fd(), file_size as i64, Whence::SeekSet)?;
            }

            // Feed existing encrypted data into HMAC calculation until EOF.
            // Read from the encrypted file starting after the header.
            // Here the last partial block is already stripped.
            #[allow(clippy::cast_possible_wrap)]
            lseek64(enc_fd.as_raw_fd(), header_size as i64, Whence::SeekSet)?;
            loop {
                let n = retry_on_eintr(|| {
                    splice(
                        &enc_fd,
                        None,
                        &pipe_wr_mac,
                        None,
                        *PIPE_BUF_ALG,
                        SpliceFFlags::empty(),
                    )
                })?;
                if n == 0 {
                    break;
                }

                let mut ncopy = n;
                while ncopy > 0 {
                    let n = retry_on_eintr(|| {
                        splice(
                            &pipe_rd_mac,
                            None,
                            &sock_mac,
                            None,
                            ncopy,
                            SpliceFFlags::SPLICE_F_MORE,
                        )
                    })?;
                    if n == 0 {
                        return Err(Errno::EBADMSG);
                    }
                    ncopy -= n;
                }
            }
        } else {
            // Non-append mode: overwrite the file.

            // Reset crypt_fd to the beginning.
            lseek64(crypt_fd.as_raw_fd(), 0, Whence::SeekSet)?;

            if file_size > 0 {
                // Remove previous content,
                // SAFETY: wipe IV to avoid reuse.
                retry_on_eintr(|| ftruncate64(enc_fd.as_raw_fd(), 0))?;
                lseek64(enc_fd.as_raw_fd(), 0, Whence::SeekSet)?;
            }

            // Write file magic and IV to the beginning of the file.
            // Leave gap for HMAC to write later.
            // SAFETY: We need the write(2) system call to write file
            // magic, HMAC and IV to the file so our seccomp filter
            // unfortunately allows it. We do our best by only allowing
            // writes up the HMAC size, which is 32 bytes. Arguably,
            // pulling a BROP with only 32 bytes of buffer-space allowed
            // to transfer the binary over a socket would be really
            // tedious.
            // Alternatively writing the HMAC & IV to xattrs would be a
            // dangerous (think backups stripping xattrs), and
            // relatively less portable workaround.
            let buf = &CRYPT_MAGIC;
            let mut nwrite = 0;
            while nwrite < buf.len() {
                #[allow(clippy::arithmetic_side_effects)]
                match write(&enc_fd, &buf[nwrite..]) {
                    Ok(0) => return Err(Errno::EINVAL),
                    Ok(n) => nwrite += n,
                    Err(Errno::EINTR) => continue,
                    Err(errno) => return Err(errno),
                }
            }

            // Move the file offset forward by HMAC_TAG_SIZE to leave
            // space for the HMAC tag. This space is going to be a
            // hole until we write back at the end, see lseek(2).
            // lseek64(enc_fd.as_raw_fd(), HMAC_TAG_SIZE as i64, Whence::SeekCur)?;
            // SAFETY: ^^ This is not portable, instead we zero it out!
            // Write HMAC placeholder (zeroed out) to reserve space for HMAC tag.
            let hmac_placeholder = [0u8; HMAC_TAG_SIZE];
            let mut nwrite = 0;
            while nwrite < hmac_placeholder.len() {
                #[allow(clippy::arithmetic_side_effects)]
                match write(&enc_fd, &hmac_placeholder[nwrite..]) {
                    Ok(0) => return Err(Errno::EINVAL),
                    Ok(n) => nwrite += n,
                    Err(Errno::EINTR) => continue,
                    Err(errno) => return Err(errno),
                }
            }

            // Write the IV to the file.
            let buf = iv.as_ref();
            let mut nwrite = 0;
            while nwrite < buf.len() {
                #[allow(clippy::arithmetic_side_effects)]
                match write(&enc_fd, &buf[nwrite..]) {
                    Ok(0) => return Err(Errno::EINVAL),
                    Ok(n) => nwrite += n,
                    Err(Errno::EINTR) => continue,
                    Err(errno) => return Err(errno),
                }
            }
        }

        // Initialize encryption socket, and set IV.
        let sock_enc = aes_ctr_init(&aes_fd, false)?;
        aes_ctr_enc(&sock_enc, &[], Some(&iv), true)?;

        // The IV is no longer needed.
        drop(iv);

        let (pipe_rd_enc, pipe_wr_enc) = pipe2(OFlag::O_CLOEXEC)?;

        // Feed plaintext via zero-copy into the kernel socket.
        let mut nflush = 0;
        loop {
            let nfeed = retry_on_eintr(|| {
                splice(
                    &crypt_fd,
                    None,
                    &pipe_wr_enc,
                    None,
                    *PIPE_BUF_ALG,
                    SpliceFFlags::empty(),
                )
            })?;
            if nfeed == 0 {
                break;
            }

            let mut ncopy = nfeed;
            while ncopy > 0 {
                let n = retry_on_eintr(|| {
                    splice(
                        &pipe_rd_enc,
                        None,
                        &sock_enc,
                        None,
                        ncopy,
                        SpliceFFlags::SPLICE_F_MORE,
                    )
                })?;
                if n == 0 {
                    return Err(Errno::EBADMSG);
                }
                ncopy -= n;
            }

            nflush += nfeed as u64;
            #[allow(clippy::cast_possible_truncation)]
            while nflush >= BLOCK_SIZE as u64 {
                let len = nflush - (nflush % BLOCK_SIZE as u64);
                let len = len as usize;
                let n = retry_on_eintr(|| {
                    splice(
                        &sock_enc,
                        None,
                        &pipe_wr_enc,
                        None,
                        len,
                        SpliceFFlags::SPLICE_F_MORE,
                    )
                })?;
                if n == 0 {
                    return Err(Errno::EBADMSG);
                }

                // Duplicate data from encryption pipe to the MAC pipe using tee(2).
                let mut ntee = n;
                while ntee > 0 {
                    let ntee_size = ntee.min(*PIPE_BUF_ALG);
                    let n_tee = retry_on_eintr(|| {
                        tee(&pipe_rd_enc, &pipe_wr_mac, ntee_size, SpliceFFlags::empty())
                    })?;
                    if n_tee == 0 {
                        return Err(Errno::EBADMSG);
                    }
                    ntee -= n_tee;
                }

                // Splice encrypted data to output file.
                let mut ncopy = n;
                while ncopy > 0 {
                    let n = retry_on_eintr(|| {
                        splice(
                            &pipe_rd_enc,
                            None,
                            &enc_fd,
                            None,
                            ncopy,
                            SpliceFFlags::empty(),
                        )
                    })?;
                    if n == 0 {
                        return Err(Errno::EBADMSG);
                    }
                    ncopy -= n;
                    nflush -= n as u64;
                }

                // Splice duplicated data to HMAC socket.
                let mut ncopy_mac = n;
                while ncopy_mac > 0 {
                    let n = retry_on_eintr(|| {
                        splice(
                            &pipe_rd_mac,
                            None,
                            &sock_mac,
                            None,
                            ncopy_mac,
                            SpliceFFlags::SPLICE_F_MORE,
                        )
                    })?;
                    if n == 0 {
                        return Err(Errno::EBADMSG);
                    }
                    ncopy_mac -= n;
                }
            }
        }

        // Finalize encryption with `false`.
        aes_ctr_enc(&sock_enc, &[], None, false)?;

        // Flush the final batch.
        #[allow(clippy::cast_possible_truncation)]
        while nflush > 0 {
            let len = nflush.min(*PIPE_BUF_ALG as u64) as usize;
            let n = retry_on_eintr(|| {
                splice(
                    &sock_enc,
                    None,
                    &pipe_wr_enc,
                    None,
                    len,
                    SpliceFFlags::empty(),
                )
            })?;
            if n == 0 {
                return Err(Errno::EBADMSG);
            }

            // Duplicate data from encryption pipe to the MAC pipe using tee(2).
            let mut ntee = n;
            while ntee > 0 {
                let ntee_size = ntee.min(*PIPE_BUF_ALG);
                let n_tee = retry_on_eintr(|| {
                    tee(&pipe_rd_enc, &pipe_wr_mac, ntee_size, SpliceFFlags::empty())
                })?;
                if n_tee == 0 {
                    return Err(Errno::EBADMSG);
                }
                ntee -= n_tee;
            }

            // Splice encrypted data to output file.
            let mut ncopy = n;
            while ncopy > 0 {
                let n = retry_on_eintr(|| {
                    splice(
                        &pipe_rd_enc,
                        None,
                        &enc_fd,
                        None,
                        ncopy,
                        SpliceFFlags::empty(),
                    )
                })?;
                if n == 0 {
                    return Err(Errno::EBADMSG);
                }
                ncopy -= n;
                nflush -= n as u64;
            }

            // Splice duplicated data to HMAC socket.
            let mut ncopy_mac = n;
            while ncopy_mac > 0 {
                let n = retry_on_eintr(|| {
                    splice(
                        &pipe_rd_mac,
                        None,
                        &sock_mac,
                        None,
                        ncopy_mac,
                        SpliceFFlags::SPLICE_F_MORE,
                    )
                })?;
                if n == 0 {
                    return Err(Errno::EBADMSG);
                }
                ncopy_mac -= n;
            }
        }

        // Finalize HMAC computation and retrieve the tag.
        // SAFETY: This is the only place where we use
        // the read(2) system call hence we allow read(2)
        // system call up to 32 bytes which is the size
        // of the HMAC.
        let hmac_tag = hmac_sha256_fini(&sock_mac)?;

        // Seek back to the position after the magic header.
        #[allow(clippy::cast_possible_wrap)]
        lseek64(
            enc_fd.as_raw_fd(),
            CRYPT_MAGIC.len() as i64,
            Whence::SeekSet,
        )?;

        // Write the HMAC tag to the file.
        let buf = hmac_tag.unsecure();
        let mut nwrite = 0;
        while nwrite < buf.len() {
            #[allow(clippy::arithmetic_side_effects)]
            match write(&enc_fd, &buf[nwrite..]) {
                Ok(0) => return Err(Errno::EINVAL),
                Ok(n) => nwrite += n,
                Err(Errno::EINTR) => continue,
                Err(errno) => return Err(errno),
            }
        }

        // All good, farewell to all OwnedFds!
        Ok(())
    }

    /// Confine Worker thread.
    #[allow(clippy::cognitive_complexity)]
    fn prepare_confine(safe_setid: bool) -> SydResult<ScmpFilterContext> {
        // Create seccomp filter with default action.
        let mut ctx = ScmpFilterContext::new_filter(ScmpAction::KillProcess)?;

        // Enforce the NO_NEW_PRIVS functionality before
        // loading the seccomp filter into the kernel.
        ctx.set_ctl_nnp(true)?;

        // DO NOT synchronize filter to all threads.
        // Main thread will confine itself.
        ctx.set_ctl_tsync(false)?;

        // We kill for bad system call and bad arch.
        ctx.set_act_badarch(ScmpAction::KillProcess)?;

        // SAFETY: Do NOT add supported architectures to the filter.
        // This ensures Syd can never run a non-native system call,
        // which we do not need at all.
        // seccomp_add_architectures(&mut ctx)?;

        // Deny open{,at,at2} with EACCES rather than KillProcess.
        // We need this because std::thread::spawn has unwanted
        // side-effects such as opening /sys/devices/system/cpu/online
        // on some architectures.
        for sysname in ["open", "openat", "openat2"] {
            match ScmpSyscall::from_name(sysname) {
                Ok(syscall) => {
                    ctx.add_rule(ScmpAction::Errno(Errno::EACCES as i32), syscall)?;
                    ctx.set_syscall_priority(syscall, syscall_priority(sysname))?;
                }
                Err(_) => {
                    info!("ctx": "allow_syscall",
                        "msg": format!("invalid or unsupported syscall {sysname}"));
                }
            }
        }

        // Allow reads/writes up to HMAC & IV size.
        // See the note in sync_file().
        for sysname in ["read", "write"] {
            match ScmpSyscall::from_name(sysname) {
                Ok(syscall) => {
                    ctx.add_rule_conditional(
                        ScmpAction::Allow,
                        syscall,
                        &[scmp_cmp!($arg2 <= HMAC_TAG_SIZE.max(IV_SIZE) as u64)],
                    )?;
                    ctx.set_syscall_priority(syscall, syscall_priority(sysname))?;
                }
                Err(_) => {
                    info!("ctx": "allow_syscall",
                        "msg": format!("invalid or unsupported syscall {sysname}"));
                }
            }
        }

        // Allow safe system calls.
        for sysname in AES_SYSCALLS {
            match ScmpSyscall::from_name(sysname) {
                Ok(syscall) => {
                    ctx.add_rule(ScmpAction::Allow, syscall)?;
                    ctx.set_syscall_priority(syscall, syscall_priority(sysname))?;
                }
                Err(_) => {
                    info!("ctx": "allow_syscall",
                        "msg": format!("invalid or unsupported syscall {sysname}"));
                }
            }
        }

        // Allow get id system calls.
        for sysname in GET_ID_SYSCALLS {
            match ScmpSyscall::from_name(sysname) {
                Ok(syscall) => {
                    ctx.add_rule(ScmpAction::Allow, syscall)?;
                    ctx.set_syscall_priority(syscall, syscall_priority(sysname))?;
                }
                Err(_) => {
                    info!("ctx": "allow_syscall",
                        "msg": format!("invalid or unsupported syscall {sysname}"));
                }
            }
        }

        if safe_setid {
            // SAFETY: Main thread confines these further.
            // As these system calls as per-process,
            // the main thread's seccomp rules will apply
            // to us even without TSYNC.
            for sysname in SET_ID_SYSCALLS {
                match ScmpSyscall::from_name(sysname) {
                    Ok(syscall) => {
                        ctx.add_rule(ScmpAction::Allow, syscall)?;
                        ctx.set_syscall_priority(syscall, syscall_priority(sysname))?;
                    }
                    Err(_) => {
                        info!("ctx": "allow_syscall",
                            "msg": format!("invalid or unsupported syscall {sysname}"));
                    }
                }
            }

            // SAFETY: cap{g,s}et is necessary to drop CAP_SET{U,G}ID after
            // changing {U,G}ID.
            // SAFETY: We cannot deny _sigreturn_(2) with SafeSetID!
            // NPTL uses reserved-signals during UID/GID change!
            for sysname in [
                "capget",
                "capset",
                "sigaction",
                "rt_sigaction",
                "sigreturn",
                "rt_sigreturn",
            ] {
                match ScmpSyscall::from_name(sysname) {
                    Ok(syscall) => {
                        ctx.add_rule(ScmpAction::Allow, syscall)?;
                        ctx.set_syscall_priority(syscall, syscall_priority(sysname))?;
                    }
                    Err(_) => {
                        info!("ctx": "allow_syscall",
                            "msg": format!("invalid or unsupported syscall {sysname}"));
                    }
                }
            }
        }

        Ok(ctx)
    }
}

#[derive(Clone)]
struct Worker {
    fd: RawFd,
    safe_setid: bool,
    epoll: Arc<Epoll>,
    cache: Arc<(PathCache, AddrCache)>,
    sandbox: Arc<RwLock<Sandbox>>,
    handlers: Arc<HandlerMap>,
    keep_alive: Option<u16>,
    should_exit: Arc<AtomicBool>,
    worker_data: Arc<WorkerData>,
    crypt_map: Option<AesMap>,
}

impl Worker {
    fn new(
        fd: RawFd,
        safe_setid: bool,
        epoll: Arc<Epoll>,
        cache: Arc<(PathCache, AddrCache)>,
        sandbox: Arc<RwLock<Sandbox>>,
        handlers: Arc<HandlerMap>,
        keep_alive: Option<u16>,
        should_exit: Arc<AtomicBool>,
        worker_data: Arc<WorkerData>,
        crypt_map: Option<AesMap>,
    ) -> Self {
        Worker {
            fd,
            safe_setid,
            epoll,
            cache,
            sandbox,
            handlers,
            keep_alive,
            should_exit,
            worker_data,
            crypt_map,
        }
    }

    fn try_spawn(self, mut confined: bool) -> Result<JoinHandle<()>, std::io::Error> {
        thread::Builder::new()
            .name("syd_emu".to_string())
            .stack_size(EMU_STACK_SIZE)
            .spawn(move || {
                // Create sentinel, that will handle graceful teardown.
                let mut sentinel = Sentinel::new(&self);

                // Set a logging panic hook. The default panic
                // hook calls system calls not permitted by emulators
                // such as getcwd(2), stat(2) etc.
                std::panic::set_hook(Box::new(|info| {
                    let err = match info.payload().downcast_ref::<&'static str>() {
                        Some(s) => *s,
                        None => match info.payload().downcast_ref::<String>() {
                            Some(s) => &**s,
                            None => "?",
                        },
                    };
                    let file = info.location().map(|l| l.file());
                    let line = info.location().map(|l| l.line());
                    error!("ctx": "panic", "err": err, "file": file, "line": line);
                }));

                // Thread successfully started, increment total worker count.
                self.worker_data.increment_worker_total();

                // Do not timeout before handling min-requests.
                let mut nreq = 0;
                loop {
                    // SAFETY: Confine if/once locked.
                    if !confined && Sandbox::locked_once() {
                        // SAFETY: Panic if we cannot confine the thread.
                        Self::prepare_confine(self.safe_setid)
                            .expect("prepare emulator thread confinement")
                            .load()
                            .expect("confine emulator thread");
                        info!("ctx": "confine_emulator_thread",
                            "msg": format!("emulator thread confined with{} SROP mitigation",
                                if self.safe_setid { "out" } else { "" }));
                        confined = true;
                    }

                    // Wait for the request to become ready as necessary.
                    // epoll_wait(2) will timeout and exit for non-core threads.
                    if self.poll(nreq).is_err() {
                        // Timeout or critical error.
                        // Decrement worker total and exit.
                        self.worker_data.decrement_worker_total();
                        break;
                    }

                    // Receive seccomp notification.
                    let request = if let Ok(request) = self.receive() {
                        request
                    } else {
                        // Critical error, decrement worker total and exit.
                        self.worker_data.decrement_worker_total();
                        break;
                    };

                    if let Some(request) = request {
                        // Mark thread busy and handle request.
                        sentinel.seccomp_id = Some(request.id);
                        self.worker_data.increment_worker_busy();
                        self.handle(request);
                        sentinel.seccomp_id = None;
                        self.worker_data.decrement_worker_busy();
                        nreq = nreq.saturating_add(1);
                    } // else process died-midway, continue.
                }
            })
    }

    fn receive(&self) -> Result<Option<ScmpNotifReq>, Errno> {
        // Receive and return request.
        // Break if file descriptor was closed.
        // Ignore rest of the errors as we cannot handle them,
        // e.g: EINTR|ENOENT: task is killed mid-way.
        match self.read() {
            Ok(request) => Ok(Some(request)),
            Err(Errno::EBADF) => Err(Errno::EBADF),
            Err(_) => Ok(None),
        }
    }

    fn read(&self) -> Result<ScmpNotifReq, Errno> {
        // Use libc::seccomp_notif rather than libseccomp_sys's.
        // The latter is opaque and required us to do a heap
        // allocation which we don't always want.
        let mut req: MaybeUninit<nix::libc::seccomp_notif> = MaybeUninit::zeroed();
        let ptr = req.as_mut_ptr() as *mut seccomp_notif;

        // SAFETY: libseccomp's wrapper allocates each call.
        // Note: EINTR means child killed by signal!
        match unsafe { seccomp_notify_receive(self.fd, ptr) } {
            0 => {
                // SAFETY: seccomp_notify_receive returned success.
                // Request is populated and accessing it is safe.
                ScmpNotifReq::from_sys(unsafe { req.assume_init() })
            }
            _ => Err(Errno::last()),
        }
    }

    #[allow(clippy::cognitive_complexity)]
    fn handle(&self, req: ScmpNotifReq) {
        let request = UNotifyEventRequest::new(
            req,
            self.fd,
            Arc::clone(&self.cache),
            Arc::clone(&self.sandbox),
            self.crypt_map.as_ref().map(Arc::clone),
        );
        let sydcall = Sydcall(req.data.syscall, req.data.arch);
        let handler = if let Some(handler) = self.handlers.get(&sydcall) {
            handler
        } else {
            unreachable!(
                "BUG: Missing hook for syscall {}!",
                i32::from(req.data.syscall)
            )
        };
        let resp = handler(&request);
        if resp.id == 0 && resp.val == 0 && resp.error == 0 && resp.flags == 0 {
            // Dummy seccomp response.
            // A previous addfd request has
            // already replied to the request.
            // Nothing left to do here.
            return;
        } else if resp.error == EOWNERDEAD {
            // EOWNERDEAD is a pseudo errno used by
            // the stat handler thread to close the
            // seccomp notify fd upon receiving the
            // "ghost" command.
            let req = request.get_request();
            crate::warn!("ctx": "enter_ghost_mode", "pid": req.pid,
                "sys": sydcall, "arch": SydArch(req.data.arch), "args": req.data.args,
                "src": proc_mmap(req.pid(), req.data.arch, req.data.instr_pointer));
        }
        let mut resp = nix::libc::seccomp_notif_resp {
            id: resp.id,
            val: resp.val,
            error: resp.error,
            flags: resp.flags,
        };
        // libc's data type is not opaque unlike libseccomp's.
        let ptr: *mut seccomp_notif_resp = std::ptr::addr_of_mut!(resp) as *mut seccomp_notif_resp;

        // SAFETY:
        // 1. libseccomp's version allocates needlessly
        // 2. Nothing we can do on errors, EINTR means child dead.
        unsafe { seccomp_notify_respond(self.fd, ptr) };

        // See above.
        if resp.error == EOWNERDEAD {
            // Note, threads blocked on epoll_wait will not
            // wake up even if we close the epoll fd or
            // delete the seccomp fd from epoll wait-list here.
            // That said, they'll never ever wake up again,
            // and therefore will not consume system resources.
            let _ = close(self.fd);

            // Inform the monitor thread to exit.
            self.should_exit.store(true, Ordering::Relaxed);
        }
    }

    fn poll(&self, num_requests: u64) -> Result<(), Errno> {
        let timeout = if let Some(keep_alive) = self.keep_alive {
            if num_requests > EMU_MIN_REQUESTS {
                PollTimeout::from(keep_alive)
            } else {
                PollTimeout::NONE
            }
        } else {
            PollTimeout::NONE
        };

        // Wait for an event and handle EINTR.
        let mut events = [EpollEvent::empty(); 1];
        loop {
            return match self.epoll.wait(&mut events, timeout) {
                Ok(0) if self.keep_alive.is_some() => Err(Errno::ETIMEDOUT),
                Ok(0) | Err(Errno::EINTR) => continue,
                Ok(_) if events[0].events().contains(EpollFlags::EPOLLIN) => Ok(()),
                Ok(_) => Err(Errno::EBADF), // POLLHUP!
                Err(errno) => Err(errno),
            };
        }
    }

    /// Confine Worker thread.
    #[allow(clippy::cognitive_complexity)]
    fn prepare_confine(safe_setid: bool) -> SydResult<ScmpFilterContext> {
        // Create seccomp filter with default action.
        let mut ctx = ScmpFilterContext::new_filter(ScmpAction::KillProcess)?;

        // Enforce the NO_NEW_PRIVS functionality before
        // loading the seccomp filter into the kernel.
        ctx.set_ctl_nnp(true)?;

        // DO NOT synchronize filter to all threads.
        // Main thread will confine itself.
        ctx.set_ctl_tsync(false)?;

        // We kill for bad system call and bad arch.
        ctx.set_act_badarch(ScmpAction::KillProcess)?;

        // SAFETY: Do NOT add supported architectures to the filter.
        // This ensures Syd can never run a non-native system call,
        // which we do not need at all.
        // seccomp_add_architectures(&mut ctx)?;

        // Allow safe system calls.
        for sysname in EMU_SYSCALLS {
            match ScmpSyscall::from_name(sysname) {
                Ok(syscall) => {
                    ctx.add_rule(ScmpAction::Allow, syscall)?;
                    ctx.set_syscall_priority(syscall, syscall_priority(sysname))?;
                }
                Err(_) => {
                    info!("ctx": "allow_syscall",
                        "msg": format!("invalid or unsupported syscall {sysname}"));
                }
            }
        }

        // Allow get id system calls.
        for sysname in GET_ID_SYSCALLS {
            match ScmpSyscall::from_name(sysname) {
                Ok(syscall) => {
                    ctx.add_rule(ScmpAction::Allow, syscall)?;
                    ctx.set_syscall_priority(syscall, syscall_priority(sysname))?;
                }
                Err(_) => {
                    info!("ctx": "allow_syscall",
                        "msg": format!("invalid or unsupported syscall {sysname}"));
                }
            }
        }

        if safe_setid {
            // SAFETY: Main thread confines these further.
            // As these system calls as per-process,
            // the main thread's seccomp rules will apply
            // to us even without TSYNC.
            for sysname in SET_ID_SYSCALLS {
                match ScmpSyscall::from_name(sysname) {
                    Ok(syscall) => {
                        ctx.add_rule(ScmpAction::Allow, syscall)?;
                        ctx.set_syscall_priority(syscall, syscall_priority(sysname))?;
                    }
                    Err(_) => {
                        info!("ctx": "allow_syscall",
                            "msg": format!("invalid or unsupported syscall {sysname}"));
                    }
                }
            }

            // SAFETY: cap{g,s}et is necessary to drop CAP_SET{U,G}ID after
            // changing {U,G}ID.
            // SAFETY: We cannot deny _sigreturn_(2) with SafeSetID!
            // NPTL uses reserved-signals during UID/GID change!
            for sysname in [
                "capget",
                "capset",
                "sigaction",
                "rt_sigaction",
                "sigreturn",
                "rt_sigreturn",
            ] {
                match ScmpSyscall::from_name(sysname) {
                    Ok(syscall) => {
                        ctx.add_rule(ScmpAction::Allow, syscall)?;
                        ctx.set_syscall_priority(syscall, syscall_priority(sysname))?;
                    }
                    Err(_) => {
                        info!("ctx": "allow_syscall",
                            "msg": format!("invalid or unsupported syscall {sysname}"));
                    }
                }
            }
        }

        Ok(ctx)
    }
}

/// Type that exists to manage worker exit on panic.
///
/// This type is constructed once per `Worker` and implements `Drop` to
/// handle proper worker exit in case the worker panics when executing
/// the current task or anywhere else in its work loop. If the
/// `Sentinel` is dropped at the end of the worker's work loop and the
/// current thread is panicking, handle worker exit the same way as if
/// the task completed normally (if the worker panicked while executing
/// a submitted task) then clone the worker and start it with an initial
/// task of `None`.
struct Sentinel<'s> {
    seccomp_id: Option<u64>,
    worker_ref: &'s Worker,
}

impl Sentinel<'_> {
    fn new(worker_ref: &Worker) -> Sentinel<'_> {
        Sentinel {
            seccomp_id: None,
            worker_ref,
        }
    }

    #[allow(clippy::arithmetic_side_effects)]
    fn deny_syscall(&self, seccomp_id: u64, errno: Errno) {
        let mut resp = nix::libc::seccomp_notif_resp {
            id: seccomp_id,
            val: 0,
            error: -(errno as i32),
            flags: 0,
        };
        // libc's data type is not opaque unlike libseccomp's.
        let ptr: *mut seccomp_notif_resp = std::ptr::addr_of_mut!(resp) as *mut seccomp_notif_resp;

        // SAFETY:
        // 1. libseccomp's version allocates needlessly
        // 2. Nothing we can do on errors, EINTR means child dead.
        unsafe { seccomp_notify_respond(self.worker_ref.fd, ptr) };
    }
}

impl Drop for Sentinel<'_> {
    fn drop(&mut self) {
        if thread::panicking() {
            if let Some(seccomp_id) = self.seccomp_id {
                // Busy thread panicked.
                // SAFETY: Deny syscall in progress!
                self.deny_syscall(seccomp_id, Errno::EACCES);
                self.worker_ref.worker_data.decrement_both();
            } else {
                // Idle thread panicked.
                self.worker_ref.worker_data.decrement_worker_total();
            }
        }
    }
}

const WORKER_BUSY_MASK: u32 = 0x0000FFFF;
const INCREMENT_TOTAL: u32 = 0x00010000;
const INCREMENT_BUSY: u32 = 0x00000001;

/// 1. Struct containing data shared between workers.
/// 2. Struct that stores and handles an `AtomicU32` that stores the
///    total worker count in the higher half of bits and the busy worker
///    count in the lower half of bits. This allows to to increment /
///    decrement both counters in a single atomic operation.
#[derive(Default)]
struct WorkerData(AtomicU32);

impl WorkerData {
    fn increment_both(&self) -> (u16, u16) {
        let old_val = self
            .0
            .fetch_add(INCREMENT_TOTAL | INCREMENT_BUSY, Ordering::Relaxed);
        Self::split(old_val)
    }

    fn decrement_both(&self) -> (u16, u16) {
        let old_val = self
            .0
            .fetch_sub(INCREMENT_TOTAL | INCREMENT_BUSY, Ordering::Relaxed);
        Self::split(old_val)
    }

    fn increment_worker_total(&self) -> u16 {
        let old_val = self.0.fetch_add(INCREMENT_TOTAL, Ordering::Relaxed);
        Self::total(old_val)
    }

    #[allow(dead_code)]
    fn decrement_worker_total(&self) -> u16 {
        let old_val = self.0.fetch_sub(INCREMENT_TOTAL, Ordering::Relaxed);
        Self::total(old_val)
    }

    fn increment_worker_busy(&self) -> u16 {
        let old_val = self.0.fetch_add(INCREMENT_BUSY, Ordering::Relaxed);
        Self::busy(old_val)
    }

    fn decrement_worker_busy(&self) -> u16 {
        let old_val = self.0.fetch_sub(INCREMENT_BUSY, Ordering::Relaxed);
        Self::busy(old_val)
    }

    fn get_total_count(&self) -> u16 {
        Self::total(self.0.load(Ordering::Relaxed))
    }

    fn get_busy_count(&self) -> u16 {
        Self::busy(self.0.load(Ordering::Relaxed))
    }

    #[inline]
    fn split(val: u32) -> (u16, u16) {
        let total_count = (val >> 16) as u16;
        let busy_count = (val & WORKER_BUSY_MASK) as u16;
        (total_count, busy_count)
    }

    #[inline]
    fn total(val: u32) -> u16 {
        (val >> 16) as u16
    }

    #[inline]
    fn busy(val: u32) -> u16 {
        (val & WORKER_BUSY_MASK) as u16
    }
}
